from typing import Any

import jinja2

from brigid.jinja2_render.templates import get_jinjaglobals
from brigid.plugins.core import jinjaglobals
from brigid.plugins.core.settings import settings
from brigid.plugins.plugin import Plugin


class CorePlugin(Plugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def templates_loader(self) -> jinja2.BaseLoader:
        return jinja2.FileSystemLoader(settings.templates, followlinks=True)

    def jinjaglobals(self) -> tuple[dict[str, Any], dict[str, Any]]:
        return get_jinjaglobals(jinjaglobals)


plugin = CorePlugin(slug="core")



from brigid.core import logging
from brigid.plugins.plugin import Plugin
from brigid.plugins.settings import settings

logger = logging.get_module_logger()

_plugins: list["Plugin"] | None = None
_plugins_dict: dict[str, "Plugin"] | None = None


def plugins() -> list[Plugin]:
    global _plugins

    if _plugins:
        return _plugins

    _plugins = []

    for plugin_path in settings.plugins:
        logger.info("loading_plugin", plugin=plugin_path)

        module_name, plugin_name = plugin_path.rsplit(":", 1)

        try:
            module = __import__(module_name, fromlist=[plugin_name])
        except BaseException as e:
            logger.exception("error_loading_plugin", plugin=plugin_path, exc_info=e)
            continue

        try:
            plugin = getattr(module, plugin_name)
        except BaseException as e:
            logger.exception("error_getting_plugin", plugin=plugin_path, exc_info=e)
            continue

        _plugins.append(plugin)

        logger.info("plugin_loaded", slug=plugin.slug)

    return _plugins


def plugins_dict() -> dict[str, Plugin]:
    global _plugins_dict

    if _plugins_dict:
        return _plugins_dict

    _plugins_dict = {plugin.slug: plugin for plugin in plugins()}

    return _plugins_dict


def get_plugin(slug: str) -> Plugin | None:
    return plugins_dict().get(slug)



import datetime
import enum
from typing import Annotated

from pydantic import Field

from brigid.core.entities import BaseEntity
from brigid.library.entities import ArticleType


class RenderFormat(enum.StrEnum):
    markdown = "markdown"
    html = "html"


Language = Annotated[str, Field(description="Language code of the content")]
Slug = Annotated[str, Field(description="Slug identifier of the blog post")]
RenderFormatType = Annotated[
    RenderFormat, Field(description="Format of the rendered content", default=RenderFormat.html)
]
PageNumber = Annotated[int, Field(description="Page number for pagination, starts with 1", default=1, ge=1)]
RequiredTags = Annotated[
    set[str], Field(description="Set of required tags for filtering blog posts", default_factory=set)
]
ExcludedTags = Annotated[
    set[str], Field(description="Set of excluded tags for filtering blog posts", default_factory=set)
]
PostTitle = Annotated[str, Field(description="Title of the blog post")]
TagInfos = Annotated[list["TagInfo"], Field(description="List of tags with their counts for filtered posts")]


class PostMeta(BaseEntity):
    """Metadata information about a blog post."""

    published_at: Annotated[datetime.datetime, Field(description="Publication date and time of the blog post")]
    language: Language
    translated_into: Annotated[set[Language], Field(description="List of languages the blog post is translated into")]
    slug: Slug
    seo_description: Annotated[str, Field(description="SEO description of the blog post")]
    seo_image: Annotated[str | None, Field(description="SEO image URL of the blog post")]
    tags: TagInfos
    series: Annotated[str | None, Field(description="Series identifier if the blog post is part of a series")]
    type: Annotated[ArticleType, Field(description="Type of the article, e.g., post or page")]
    http_url: Annotated[str, Field(description="HTTP URL of the blog post")]


class PostInfo(BaseEntity):
    meta: PostMeta
    title: PostTitle
    intro_format: RenderFormatType
    intro_body: Annotated[str, Field(description="Introductory content of the blog post")]
    has_more: Annotated[bool, Field(description="Indicates if the blog post has a 'read more' section")]


class Post(BaseEntity):
    meta: PostMeta
    title: PostTitle
    body_format: RenderFormatType
    body: Annotated[str, Field(description="Full content of the blog post")]


class FilteredPosts(BaseEntity):
    total_posts: Annotated[int, Field(description="Total number of posts available after filtering")]
    total_pages: Annotated[int, Field(description="Total number of pages available after filtering")]
    page_number: PageNumber
    posts: Annotated[list[PostInfo], Field(description="List of blog posts")]
    required_tags: RequiredTags
    excluded_tags: ExcludedTags
    tags: TagInfos


class TagInfo(BaseEntity):
    tag: Annotated[str, Field(description="Tag id")]
    name: Annotated[str, Field(description="Human-readable name of the tag")]
    count: Annotated[int, Field(description="Number of posts associated with the tag")]



from typing import Any

import pydantic
from pymdownx.blocks import BlocksExtension

from brigid.core.entities import BaseEntity
from brigid.markdown_render.processors.toml_block import TomlBlock


class YouTubeModel(BaseEntity):
    caption: str
    id: str
    css_class: str | None = None

    model_config = pydantic.ConfigDict(frozen=False)

    @property
    def url(self) -> str:
        return f"https://www.youtube.com/embed/{self.id}"


class YouTubeBlock(TomlBlock):
    NAME = "brigid-youtube"
    models = YouTubeModel
    root_tag = "figure"
    template = "theme/blocks/youtube.html.j2"

    def root_css_classes(self, data: Any) -> list[str]:
        classes = ["brigid-youtube", data.css_class]
        return [x for x in classes if x]


class YouTubeBlockExtension(BlocksExtension):

    def extendMarkdownBlocks(self, md, block_mgr):
        block_mgr.register(YouTubeBlock, self.getConfigs())



from brigid.cli.application import app  # noqa: F401
from brigid.cli.commands import configs  # noqa: F401
from brigid.cli.commands import static  # noqa: F401
from brigid.cli.commands import templates  # noqa: F401
from brigid.cli.commands import validate  # noqa: F401

app()



import uuid

from brigid.library.storage import storage
from brigid.library.tests import make as library_make
from brigid.validation.page_validators import page_has_correct_series_tags, page_has_correct_tags, page_is_rendered


def test_some_tags_are_in_configs() -> None:
    """To not test in every test that tags are in configs"""
    for language in storage.get_site().languages.keys():
        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())
    assert allowed_tags


class TestPageIsRendered:

    def test_empty_body(self) -> None:
        page = library_make.page(body="")
        assert page_is_rendered(page) == []

    def test_no_errors(self) -> None:
        page = library_make.page(body="some text")
        assert page_is_rendered(page) == []

    def test_has_errors(self) -> None:
        body = """
/// brigid-images
x = "y"
///
"""

        page = library_make.page(body=body)
        assert page_is_rendered(page) != []


class TestPageHasCorrectTags:

    def test_empty_tags(self) -> None:
        page = library_make.page(body="", tags=[])
        assert page_has_correct_tags(page) == []

    def test_no_errors(self) -> None:
        language = "en"

        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())

        page = library_make.page(language=language, body="", tags=allowed_tags)
        assert page_has_correct_tags(page) == []

    def test_has_errors(self) -> None:
        language = "en"

        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())

        bad_tag_1 = uuid.uuid4().hex
        bad_tag_2 = uuid.uuid4().hex

        allowed_tags.extend([bad_tag_1, bad_tag_2])

        page = library_make.page(language=language, body="", tags=allowed_tags)

        errors = page_has_correct_tags(page)

        assert len(errors) == 2

        assert (bad_tag_1 in errors[0].message and bad_tag_2 in errors[1].message) or (
            bad_tag_1 in errors[1].message and bad_tag_2 in errors[0].message
        )


class TestPageHasCorrectSeriesTags:

    def test_empty_series(self) -> None:
        language = "en"

        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())

        page = library_make.page(body="", tags=allowed_tags, series=None)

        assert page_has_correct_series_tags(page) == []

    def test_ok_series(self) -> None:
        language = "en"

        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())

        page = library_make.page(body="", tags=allowed_tags, series=allowed_tags[0])

        assert page_has_correct_series_tags(page) == []

    def test_wrong_series(self) -> None:
        language = "en"

        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())

        bad_tag = uuid.uuid4().hex

        page = library_make.page(body="", tags=allowed_tags + [bad_tag], series=bad_tag)

        errors = page_has_correct_series_tags(page)

        assert len(errors) == 1

        assert bad_tag in errors[0].message

    def test_series_not_in_page_tags(self) -> None:
        language = "en"

        allowed_tags = list(storage.get_site().languages[language].tags_translations.keys())

        page = library_make.page(body="", tags=allowed_tags[1:], series=allowed_tags[0])

        errors = page_has_correct_series_tags(page)

        assert len(errors) == 1

        assert allowed_tags[0] in errors[0].message



import traceback
import xml.etree.ElementTree as etree  # noqa: S405
from typing import Any

import pydantic
import toml
from pymdownx.blocks.block import Block

from brigid.markdown_render.context import render_context

_renderer = None


class TomlBlock(Block):
    NAME: str = NotImplemented
    models: type[pydantic.BaseModel] = NotImplemented
    root_tag: str = "div"
    template: str = NotImplemented

    def root_css_classes(self, data: Any) -> list[str]:
        return []

    def on_create(self, parent):
        return etree.SubElement(parent, self.root_tag)

    def on_markdown(self) -> str:
        return "raw"

    def on_end(self, block: etree.Element) -> None:
        try:
            if block.text is None:
                raise NotImplementedError("Block text is None")

            data = toml.loads(block.text)

            context = render_context.get()

            raw_model = pydantic.TypeAdapter(self.models).validate_python(data)

            model = self.process_data(raw_model)

            new_text = self.render_in_theme(
                {"data": model, "current_article": context.article, "current_page": context.page}
            )

            block.text = self.md.htmlStash.store(new_text)

            if model:
                block.set("class", " ".join(self.root_css_classes(model)))

        except Exception as e:
            new_text = f"Error: {traceback.format_exc()}".replace("\n", "<br>")

            context = render_context.get()

            context.add_error(failed_text="unknown", message=f"Error while rendering block: {e}")

    def process_data(self, data: Any) -> Any:
        return data

    def render_in_theme(self, data: Any) -> str:
        from brigid.jinja2_render.templates import render

        return render(self.template, data)



from typing import Any

import jinja2
from markupsafe import Markup

from brigid.core import logging
from brigid.jinja2_render.entities import IndexInfo, Info, MetaInfo, PageInfo
from brigid.plugins.entities import FileInfo

logger = logging.get_module_logger()


class Plugin:

    def __init__(self, slug: str):
        self.slug = slug

    def templates_loader(self) -> jinja2.BaseLoader | None:
        return None

    def jinjaglobals(self) -> tuple[dict[str, Any], dict[str, Any]]:
        return {}, {}

    def static_file_info(self, filename: str) -> FileInfo | None:  # pylint: disable=W0613
        return None

    def static_files(self) -> list[FileInfo]:
        return []

    def render_template_if_exists(  # noqa: CFQ002
        self,
        template_name: str,
        info: Info,
        meta: MetaInfo,
        index: IndexInfo | None = None,
        page: PageInfo | None = None,
        default: str = "",
    ) -> str:
        from brigid.jinja2_render.templates import render

        try:
            return Markup(
                render(
                    template_name,
                    {"info": info, "meta_info": meta, "index_info": index, "page_info": page, "plugin": self},
                )
            )
        except jinja2.TemplateNotFound:
            return default
        except BaseException as e:
            logger.exception("plugin_render_error", template=template_name, plugin=self.slug, exc_info=e)
            return str(e)

    def render_head(
        self, info: Info, meta: MetaInfo, index: IndexInfo | None = None, page: PageInfo | None = None
    ) -> str:
        return self.render_template_if_exists(
            f"{self.slug}/head.html.j2", info=info, meta=meta, index=index, page=page
        )

    def render_body_before_content(
        self, info: Info, meta: MetaInfo, index: IndexInfo | None = None, page: PageInfo | None = None
    ) -> str:
        return self.render_template_if_exists(
            f"{self.slug}/body_before_content.html.j2", info=info, meta=meta, index=index, page=page
        )

    def render_body_after_content(
        self, info: Info, meta: MetaInfo, index: IndexInfo | None = None, page: PageInfo | None = None
    ) -> str:
        return self.render_template_if_exists(
            f"{self.slug}/body_after_content.html.j2", info=info, meta=meta, index=index, page=page
        )

    def render_body_after_footer(
        self, info: Info, meta: MetaInfo, index: IndexInfo | None = None, page: PageInfo | None = None
    ) -> str:
        return self.render_template_if_exists(
            f"{self.slug}/body_after_footer.html.j2", info=info, meta=meta, index=index, page=page
        )



from collections import Counter

import fastmcp

from brigid.library.storage import storage
from brigid.mcp import domain
from brigid.mcp.entities import (
    ExcludedTags,
    FilteredPosts,
    Language,
    PageNumber,
    Post,
    RenderFormatType,
    RequiredTags,
    Slug,
    TagInfo,
)

# TODO: unify tools code with the api renderers
# TODO: should we render markdown in a special format for MCP? To support backlinks, images as resources, etc.?
# TODO: should we add an instruction about the markdown format used in the blog?
# TODO: add mcp url constructors, like with http urls?


def create_tools(mcp: fastmcp.FastMCP) -> None:  # noqa: CCR001, CFQ001
    site = storage.get_site()

    get_posts_description = "\n".join(
        [
            "Returns a filtered list of blog posts from new to old.",
            f"Returns up to {site.posts_per_page} posts. Request the next page to get more posts.",
            "Always start requesting from page 1.",
            "",
            "- required_tags: A set of tags that the blog posts must have.",
            "- excluded_tags: A set of tags that the blog posts must not have.",
            "",
            "Recomendations:",
            "",
            (
                "- Filter posts by tags gradually — add one tag at a time — "
                "in response you'll find tag counts for the tags in the filtered posts."
            ),
        ]
    )

    @mcp.tool(name="get_posts", description=get_posts_description)
    def get_posts(
        language: Language,
        page_number: PageNumber,
        required_tags: RequiredTags,
        excluded_tags: ExcludedTags,
        render_format: RenderFormatType,
    ) -> FilteredPosts:
        all_posts = storage.get_posts(language=language, require_tags=required_tags, exclude_tags=excluded_tags)

        tags_count: Counter[str] = Counter()

        for post in all_posts:
            tags_count.update(post.tags)

        posts = all_posts[site.posts_per_page * (page_number - 1) : site.posts_per_page * page_number]

        total_pages = (len(all_posts) + site.posts_per_page - 1) // site.posts_per_page

        return FilteredPosts(
            total_posts=len(all_posts),
            total_pages=total_pages,
            page_number=page_number,
            posts=[domain.create_post_info(post, render_format) for post in posts],
            required_tags=required_tags,
            excluded_tags=excluded_tags,
            tags=domain.create_tag_infos(language, tags_count),
        )

    get_post_description = "\n".join(
        [
            "Returns the full content of a blog post identified by its slug in the specified language.",
            "",
            "Recommendations:",
            "",
            "- Prefer `html` as the render format when you need working links to other posts or resources.",
            (
                "- Prefer `html` as the render format when you need to display the post content "
                "'as it rendered' directly to the user."
            ),
            "- Prefer `markdown` as the render format when you need 'just this post content'.",
            "- Prefer `markdown` as the render format when you do automatic processing of the post content.",
        ]
    )

    @mcp.tool(name="get_post", description=get_post_description)
    def get_post(language: Language, slug: Slug, render_format: RenderFormatType) -> Post | None:
        if language not in site.allowed_languages:
            # TODO: send notification or error?
            return None

        if not storage.has_article(slug=slug):
            # TODO: send notification or error?
            return None

        article = storage.get_article(slug=slug)

        if language not in article.pages:
            # TODO: send notification or error?
            return None

        post = storage.get_page(id=article.pages[language])

        return domain.create_post(post, render_format)

    get_tags_description = "\n".join(
        [
            (
                "Returns a list of all tags used in blog posts for the specified language, "
                "along with the count of posts associated with each tag."
            ),
            "",
            "Recommendations:",
            "",
            (
                "- Use this tool when the user requested information about specific topics: "
                "get all tags -> choose relevant tags -> get posts with these tags."
            ),
        ]
    )

    @mcp.tool(name="get_tags", description=get_tags_description)
    def get_tags(language: Language) -> list[TagInfo]:
        all_posts = storage.get_posts(language=language, require_tags=(), exclude_tags=())

        tags_count: Counter[str] = Counter()

        for post in all_posts:
            tags_count.update(post.tags)

        return domain.create_tag_infos(language, tags_count)



from brigid.domain.text import capitalize_first
from brigid.library.entities import Page
from brigid.library.storage import storage


def page_title(page: Page, short: bool) -> str:

    if short:
        return capitalize_first(page.title)

    if page.series:
        series_title = storage.get_site().languages[page.language].tags_translations[page.series]
        return f"{capitalize_first(series_title)}: {capitalize_first(page.title)}"

    return capitalize_first(page.title)



import copy
from typing import Any
from unittest import mock

import pytest

from brigid.domain.urls import (
    UrlsAuthor,
    UrlsBase,
    UrlsFeedsAtom,
    UrlsPost,
    UrlsRoot,
    UrlsSiteMapFull,
    UrlsTags,
    normalize_url,
)
from brigid.library.storage import storage

base_url = "http://0.0.0.0:8000"


class TestNormalizeUrl:

    @pytest.mark.parametrize("url_in", ["", "example.com", "example.com/////", "child.example.com//"])
    def test_wrong_url(self, url_in):
        with pytest.raises(NotImplementedError):
            normalize_url(url_in)

    @pytest.mark.parametrize(
        "url_in,url_out",
        [
            ("https://example.com", "https://example.com"),
            ("https://example.com/", "https://example.com"),
            ("https://example.com//", "https://example.com"),
            ("https://example.com/////", "https://example.com"),
            ("https://child.example.com//", "https://child.example.com"),
            ("https://example.com/path/b/c", "https://example.com/path/b/c"),
            ("https://example.com/path/b/c/", "https://example.com/path/b/c"),
            ("https://example.com//path///b///c", "https://example.com/path/b/c"),
            ("https://example.com/path/b/c?", "https://example.com/path/b/c"),
            ("http://example.com//path///b///c", "http://example.com/path/b/c"),
        ],
    )
    def test_normalization(self, url_in: str, url_out: str):
        assert normalize_url(url_in) == url_out


class _TestUrlsBase:

    base_language = "en"
    alt_language = "ru"

    @pytest.fixture
    def url(self):
        return self._consruct_url()

    def _consruct_url(self) -> Any:
        raise NotImplementedError()

    def test_base_initialized(self, url: UrlsBase) -> None:
        assert url.language == self.base_language

    def test_url_method_redefined(self, url: UrlsBase) -> None:
        assert isinstance(url.url(), str)

    def test_file_url(self, url: UrlsBase) -> None:
        with pytest.raises(NotImplementedError):
            url.file_url("")

    def test_to_root(self, url: UrlsBase) -> None:
        assert url.to_root() == UrlsRoot(language=self.base_language)

    def test_to_feeds_atom(self, url: UrlsBase) -> None:
        assert url.to_feeds_atom() == UrlsFeedsAtom(language=self.base_language)

    def test_to_author(self, url: UrlsBase) -> None:
        assert url.to_author() == UrlsAuthor(language=self.base_language)

    def test_to_post(self, url: UrlsBase) -> None:
        assert url.to_post("xxx") == UrlsPost(language=self.base_language, slug="xxx")

    def test_to_filter(self, url: UrlsBase) -> None:
        filter_url = url.to_filter(page=13, require=("a", "b"), exclude=("c", "d", "e"))

        expected_filter_url = UrlsTags(
            language=self.base_language,
            page=13,
            required_tags=("a", "b"),
            excluded_tags=("c", "d", "e"),
        )

        assert filter_url == expected_filter_url

    def test_language(self, url: UrlsBase) -> None:
        expected_url = copy.deepcopy(url)
        expected_url.language = self.alt_language

        assert url.to_language(self.alt_language) == expected_url

    def test_site_map_full(self, url: UrlsBase) -> None:
        assert url.to_site_map_full() == UrlsSiteMapFull(language=self.base_language)

    def test_noindex(self, url: UrlsBase) -> None:
        assert not url.is_noindex()


class TestUrlsBase(_TestUrlsBase):

    def _consruct_url(self) -> UrlsBase:
        return UrlsBase(language=self.base_language)

    def test_url_method_redefined(self, url: UrlsBase) -> None:
        with pytest.raises(NotImplementedError):
            url.url()


class TestUrlsRoot(_TestUrlsBase):

    def _consruct_url(self) -> UrlsRoot:
        return UrlsRoot(language=self.base_language)

    def test_url_method_redefined(self, url: UrlsBase) -> None:
        assert url.url() == f"{base_url}/{self.base_language}"


class TestUrlsAuthor(_TestUrlsBase):

    def _consruct_url(self) -> UrlsAuthor:
        return UrlsAuthor(language=self.base_language)

    def test_url_method_redefined(self, url: UrlsBase) -> None:
        assert url.url() == f"{base_url}/{self.base_language}/about"


class TestUrlsFeedsAtom(_TestUrlsBase):

    def _consruct_url(self) -> UrlsFeedsAtom:
        return UrlsFeedsAtom(language=self.base_language)

    def test_url_method_redefined(self, url: UrlsBase) -> None:
        assert url.url() == f"{base_url}/{self.base_language}/feeds/atom"


class TestUrlsPost(_TestUrlsBase):

    def _consruct_url(self) -> UrlsPost:
        return UrlsPost(language=self.base_language, slug="some-slug")

    def test_initialized(self, url: UrlsPost) -> None:
        assert url.slug == "some-slug"

    def test_url_method_redefined(self, url: UrlsPost) -> None:  # type: ignore
        assert url.url() == f"{base_url}/{self.base_language}/posts/some-slug"

    def test_file_url(self, url: UrlsPost) -> None:  # type: ignore
        filepath = "images/some-image.png"
        assert url.file_url(filepath) == f"{base_url}/static/posts/some-slug/{filepath}"


class TestUrlsTags(_TestUrlsBase):

    def _consruct_url(self) -> UrlsTags:
        return UrlsTags(language=self.base_language, page=13, required_tags=("a", "d"), excluded_tags=("c", "b", "e"))

    def test_initialized(self, url: UrlsTags) -> None:
        assert url.page == 13
        assert url.required_tags == {"a", "d"}
        assert url.excluded_tags == {"c", "b", "e"}
        assert url.selected_tags == {"a", "b", "c", "d", "e"}

    def test_url_method_redefined(self, url: UrlsTags) -> None:  # type: ignore
        assert url.url() == f"{base_url}/{self.base_language}/tags/a/-b/-c/d/-e/13"

    def test_url_method__empty(self) -> None:
        url = UrlsTags(language=self.base_language, page=1, required_tags=(), excluded_tags=())

        assert url.url() == f"{base_url}/{self.base_language}"

    def test_url_method__only_required(self) -> None:
        url = UrlsTags(language=self.base_language, page=1, required_tags=("a", "b"), excluded_tags=())

        assert url.url() == f"{base_url}/{self.base_language}/tags/a/b"

    def test_url_method__only_excluded(self) -> None:
        url = UrlsTags(language=self.base_language, page=1, required_tags=(), excluded_tags=("a", "b"))

        assert url.url() == f"{base_url}/{self.base_language}/tags/-a/-b"

    def test_url_method__only_page(self) -> None:
        url = UrlsTags(language=self.base_language, page=13, required_tags=(), excluded_tags=())

        assert url.url() == f"{base_url}/{self.base_language}/tags/13"

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_is_same_index(self, url: UrlsTags) -> None:
        assert url._is_same_index(url)
        assert url._is_same_index(url.move_page(1))
        assert url._is_same_index(url.move_page(-1))

        assert not url._is_same_index(url.require("f"))
        assert not url._is_same_index(url.exclude("f"))
        assert not url._is_same_index(url.remove("a"))
        assert not url._is_same_index(url.remove("c"))
        assert not url._is_same_index(url.to_language(self.alt_language))

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_is_prev_to(self, url: UrlsTags) -> None:
        assert not url.is_prev_to(url)
        assert not url.is_prev_to(url.move_page(-1))
        assert url.is_prev_to(url.move_page(1))

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_is_next_to(self, url: UrlsTags) -> None:
        assert not url.is_next_to(url)
        assert url.is_next_to(url.move_page(-1))
        assert not url.is_next_to(url.move_page(1))

    def test_first_page(self, url: UrlsTags) -> None:
        assert url.first_page() == UrlsTags(
            language=self.base_language, page=1, required_tags=url.required_tags, excluded_tags=url.excluded_tags
        )

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_last_page(self, url: UrlsTags) -> None:
        assert url.last_page() == UrlsTags(
            language=self.base_language, page=100500, required_tags=url.required_tags, excluded_tags=url.excluded_tags
        )

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_move_page(self, url: UrlsTags) -> None:
        assert url.move_page(3) == UrlsTags(
            language=self.base_language,
            page=url.page + 3,
            required_tags=url.required_tags,
            excluded_tags=url.excluded_tags,
        )

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_move_page__too_low(self, url: UrlsTags) -> None:
        with pytest.raises(NotImplementedError):
            url.move_page(-10005000)

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_move_page__too_big(self, url: UrlsTags) -> None:
        with pytest.raises(NotImplementedError):
            url.move_page(10005000)

    def test_require(self, url: UrlsTags) -> None:
        assert url.require("f") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d", "f"), excluded_tags=("c", "b", "e")
        )

    def test_require__move(self, url: UrlsTags) -> None:
        assert url.require("b") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d", "b"), excluded_tags=("c", "e")
        )

    def test_require__duplicated(self, url: UrlsTags) -> None:
        assert url.require("a") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d"), excluded_tags=("c", "b", "e")
        )

    def test_exclude(self, url: UrlsTags) -> None:
        assert url.exclude("f") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d"), excluded_tags=("c", "b", "e", "f")
        )

    def test_exclude__move(self, url: UrlsTags) -> None:
        assert url.exclude("d") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a",), excluded_tags=("c", "b", "d", "e")
        )

    def test_exclude__duplicated(self, url: UrlsTags) -> None:
        assert url.exclude("b") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d"), excluded_tags=("c", "b", "e")
        )

    def test_remove__no_tag(self, url: UrlsTags) -> None:
        assert url.remove("f") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d"), excluded_tags=("c", "b", "e")
        )

    def test_remove__from_required(self, url: UrlsTags) -> None:
        assert url.remove("d") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a",), excluded_tags=("c", "b", "e")
        )

    def test_remove__from_excluded(self, url: UrlsTags) -> None:
        assert url.remove("b") == UrlsTags(
            language=self.base_language, page=1, required_tags=("a", "d"), excluded_tags=("c", "e")
        )

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_noindex(self, url: UrlsTags) -> None:  # type: ignore
        assert url.selected_tags
        assert url.is_noindex()
        assert url.move_page(1).is_noindex()
        assert url.move_page(-1).is_noindex()

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_no_noindex_for_basic_pages_list(self) -> None:
        url = UrlsTags(language=self.base_language, page=10, required_tags=(), excluded_tags=())

        assert not url.is_noindex()
        assert not url.move_page(1).is_noindex()
        assert not url.move_page(-1).is_noindex()

    def test_total_pages_cached(self, url: UrlsTags) -> None:
        assert url._total_pages is None

        assert isinstance(url.total_pages, int)

        assert url._total_pages is not None
        assert url._total_pages == url.total_pages

    def test_get_total_pages(self) -> None:
        url = UrlsTags(language=self.base_language, page=1, required_tags=(), excluded_tags=())

        all_pages = storage.get_posts(language=self.base_language, require_tags=(), exclude_tags=())
        posts_per_page = storage.get_site().posts_per_page

        pages = (len(all_pages) + posts_per_page - 1) // posts_per_page

        assert pages == url.total_pages == url.move_page(1).total_pages

    def test_get_total_pages__requied_filter(self) -> None:
        url = UrlsTags(language=self.base_language, page=1, required_tags=("basic",), excluded_tags=())

        whole_pages = storage.get_posts(language=self.base_language, require_tags=(), exclude_tags=())

        all_pages = storage.get_posts(language=self.base_language, require_tags=("basic",), exclude_tags=())

        assert len(whole_pages) != len(all_pages)

        posts_per_page = storage.get_site().posts_per_page

        pages = (len(all_pages) + posts_per_page - 1) // posts_per_page

        assert pages == url.total_pages

    def test_get_total_pages__exclued_filter(self) -> None:
        url = UrlsTags(language=self.base_language, page=1, required_tags=(), excluded_tags=("basic",))

        whole_pages = storage.get_posts(language=self.base_language, require_tags=(), exclude_tags=())

        all_pages = storage.get_posts(language=self.base_language, require_tags=(), exclude_tags=("basic",))

        assert len(whole_pages) != len(all_pages)

        posts_per_page = storage.get_site().posts_per_page

        pages = (len(all_pages) + posts_per_page - 1) // posts_per_page

        assert pages == url.total_pages



import pathlib

import pydantic_settings

from brigid.core.settings import BaseSettings


class Settings(BaseSettings):
    templates: pathlib.Path = pathlib.Path(__file__).parent / "templates"

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="BRIGID_PLUGINS_CORE_")


settings = Settings()



import datetime
import importlib
import pathlib
import sys
import types


def discover_submodules(parent_module: str) -> list[types.ModuleType]:
    parent = sys.modules.get(parent_module)

    if parent is None:
        parent = importlib.import_module(parent_module)

    if not hasattr(parent, "__path__"):
        return []

    parent_dir = pathlib.Path(parent.__path__[0])

    candidates = []

    for module_path in parent_dir.glob("*.py"):
        module_name = module_path.stem
        candidates.append(f"{parent_module}.{module_name}")

    for module_path in parent_dir.glob("*/__init__.py"):
        module_name = module_path.parent.stem
        candidates.append(f"{parent_module}.{module_name}")

    child_modules = []

    for full_module_name in candidates:
        submodule = sys.modules.get(full_module_name)

        if submodule is None:
            submodule = importlib.import_module(full_module_name)

        child_modules.append(submodule)

    return child_modules


def now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


_version = None


def version() -> str:
    global _version

    if _version is not None:
        return _version

    _version = importlib.metadata.version("brigid")

    return _version



from typing import Any

import jinja2

from brigid.jinja2_render.settings import settings
from brigid.library.storage import storage
from brigid.plugins.utils import plugins


def get_jinjaglobals(module):

    filter_functions = {}
    global_functions = {}

    for func_name in dir(module):
        func = getattr(module, func_name)

        if getattr(func, "_is_jinjaglobal", False):
            global_functions[func_name] = func

        if getattr(func, "_is_jinjafilter", False):
            filter_functions[func_name] = func

    return global_functions, filter_functions


def fill_globals(environment):
    from brigid.jinja2_render import jinjaglobals

    environment.globals.update({"site": storage.get_site(), "storage": storage, "plugins": plugins()})

    for module in (jinjaglobals,):
        global_functions, filter_functions = get_jinjaglobals(module)

        environment.globals.update(global_functions)
        environment.filters.update(filter_functions)

    # process plugins after default globals, and in defined order to allow plugins to override them
    for plugin in plugins():
        global_functions, filter_functions = plugin.jinjaglobals()

        environment.globals.update(global_functions)
        environment.filters.update(filter_functions)


def environment():
    loader = jinja2.PrefixLoader(
        {plugin.slug: loader for plugin in plugins() if (loader := plugin.templates_loader()) is not None}
    )

    environment = jinja2.Environment(
        autoescape=True,
        trim_blocks=True,
        auto_reload=settings.templates_reload,
        undefined=jinja2.StrictUndefined,
        loader=loader,
        extensions=["jinja2.ext.loopcontrols"],
    )

    fill_globals(environment)

    return environment


_env = None


def initialize():
    global _env
    _env = environment()


def env() -> jinja2.Environment:
    if _env is None:
        initialize()

    assert _env is not None

    return _env


def render(template_name: str, context: dict[str, Any]) -> str:
    template = env().get_template(template_name)
    return template.render(context)



import pathlib

import pydantic_settings

from brigid.core.settings import BaseSettings


class Settings(BaseSettings):
    cache_directory: pathlib.Path | None = None

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="BRIGID_API_")


settings = Settings()



import asyncio
import pathlib
import sys

import typer

from brigid.application.application import with_app
from brigid.cli.application import app
from brigid.core import logging
from brigid.jinja2_render.templates import environment

logger = logging.get_module_logger()


templates_cli = typer.Typer()


async def run_list() -> None:
    async with with_app():
        templates = []

        env = environment()
        for template_path in env.list_templates():
            template = env.get_template(template_path)
            full_path = pathlib.Path(template.filename).resolve()
            templates.append(f"{full_path}\t->\t{template_path}")

        templates.sort()

        sys.stdout.write("\n".join(templates) + "\n")


@templates_cli.command()
def list() -> None:
    asyncio.run(run_list())


async def run_copy(destination: pathlib.Path) -> None:
    async with with_app():
        env = environment()

        if not destination.exists():
            logger.info("creating_destination_directory", path=destination)
            destination.mkdir(parents=True, exist_ok=True)

        for template_path in env.list_templates():
            template = env.get_template(template_path)
            full_path = pathlib.Path(template.filename).resolve()

            destination_file = destination / template_path
            destination_file.parent.mkdir(parents=True, exist_ok=True)

            logger.info("copying_template", source=str(full_path), destination=str(destination_file))

            with open(destination_file, "w", encoding="utf-8") as f:
                f.write(full_path.read_text(encoding="utf-8"))


@templates_cli.command()
def copy(destination: pathlib.Path) -> None:
    """Copy all Jinja2 templates to the specified destination directory.

    May be helpfull:

    - for starting a new theme
    - for collecting all html code for tailwind to generate an optimal CSS file

    """
    asyncio.run(run_copy(destination))


app.add_typer(templates_cli, name="templates", help="Manage Jinja2 templates")



import os

import fastapi
import fastmcp
from fastmcp.server.http import StarletteWithLifespan

from brigid.core import utils
from brigid.library.storage import storage
from brigid.mcp.tools import create_tools


def construct_instructions() -> str:
    site = storage.get_site()
    site_i18n = site.languages[site.default_language]

    is_multi_lingual = len(site.allowed_languages) > 1

    source = [
        f"This is a personal assistant for readers of the blog {site.url}.",
        f"The blog title '{site_i18n.title}'.",
        f"The blog description is: '{site_i18n.subtitle}'",
        f"The blog author is '{site_i18n.author}'.",
        f"The license of the blog content is {site_i18n.license}." if site_i18n.license else None,
        (
            f"The blog is multi-lingual and supports the following languages: {', '.join(site.allowed_languages)}."
            if is_multi_lingual
            else None
        ),
        "Each piece of content in the blog can be available in multiple languages." if is_multi_lingual else None,
        f"The default language of the blog is {site.default_language}." if len(site.allowed_languages) > 1 else None,
        (
            f"The content of the blog is stored in the repository: {site.content_repository}."
            if site.content_repository
            else None
        ),
        "The posts in this blog is organized by tags. Use them to find content related to specific topics.",
        "The posts in this blog connected to similar posts, like in a graph. Use this to find related content.",
        (
            "Some posts in this blog are organized into series. "
            "Use them to explore complex topics step by step and to read posts in order."
        ),
        (
            "Some posts in this blog are ogranized into collections. "
            "Use them to explore posts highlighted by the author by specific properties."
        ),
        (
            "Besides posts, the blog contains static pages. "
            "You can look on them as on the important sections/collections of the blog."
        ),
    ]

    source = [instruction for instruction in source if instruction]

    return "\n".join(source)  # type: ignore


# We create MCP instance dymanically because:
# - we need site configs that are loaded at runtime
# - we may need to construct multiple MCPs (per language) in the future
def create_mcp(app: fastapi.FastAPI) -> StarletteWithLifespan:

    site = storage.get_site()
    site_i18n = site.languages[site.default_language]

    # There are issues with continuing session after server restart
    # For example, OpenAI does not handle it well
    # TODO: try to rollback to stateful sessions in the future
    if "FASTMCP_STATELESS_HTTP" not in os.environ:
        os.environ["FASTMCP_STATELESS_HTTP"] = "1"

    # TODO: website_url, icons (fastmcp 2.14.0+)
    mcp = fastmcp.FastMCP(
        name=site_i18n.title,
        instructions=construct_instructions(),
        version=utils.version(),
        auth=None,
        strict_input_validation=False,  # allow Pydantic to be flexible
    )

    create_tools(mcp)

    mcp_app = mcp.http_app(path="/")
    app.mount("/mcp", mcp_app)

    return mcp_app



import datetime
import xml.etree.ElementTree as ET  # noqa: S405

from brigid.domain.urls import UrlsPost, UrlsRoot
from brigid.library.storage import storage

# Google does not like too old dates like 0001-01-01
# => we do not create lastmod for pages with date less than MIN_DATE
MIN_DATE = datetime.datetime(1980, 1, 1, tzinfo=datetime.timezone.utc)


def get_last_published_at(language) -> datetime.datetime | None:
    last_published_at = None

    # TODO: here we look at all posts, but we should look at all sources of blog
    #       including configs & static files
    for page in storage.get_posts(language=language):
        if last_published_at is None or page.published_at > last_published_at:
            last_published_at = page.published_at

    return last_published_at


def add_lastmod(url: ET.Element, date: datetime.datetime | None) -> None:
    if date is None:
        return

    if date < MIN_DATE:
        return

    lastmod = ET.SubElement(url, "lastmod")
    lastmod.text = date.isoformat()


def build_sitemap_xml() -> str:

    site = storage.get_site()

    attributes = {
        "xmlns": "http://www.sitemaps.org/schemas/sitemap/0.9",
        "xmlns:xhtml": "http://www.w3.org/1999/xhtml",
    }

    sitemap = ET.Element("urlset", **attributes)  # type: ignore[arg-type]

    for page in storage.all_entities():
        add_page_url(sitemap, page)

    for language in site.allowed_languages:
        add_index_url(sitemap, language)

    return ET.tostring(sitemap, encoding="unicode")


def add_index_url(sitemap, language) -> None:

    site = storage.get_site()

    url = ET.SubElement(sitemap, "url")

    root_url = UrlsRoot(language=language)

    loc = ET.SubElement(url, "loc")
    loc.text = root_url.url()

    last_published_at = get_last_published_at(language)

    add_lastmod(url, last_published_at)

    priority = ET.SubElement(url, "priority")
    priority.text = "1.0"

    if len(site.allowed_languages) > 1:
        add_language_variants(url, root_url, site.allowed_languages)


def add_page_url(sitemap, page) -> None:

    article = storage.get_article(id=page.article_id)

    url = ET.SubElement(sitemap, "url")

    page_url = UrlsPost(language=page.language, slug=article.slug)

    loc = ET.SubElement(url, "loc")
    loc.text = page_url.url()

    add_lastmod(url, page.published_at)

    priority = ET.SubElement(url, "priority")
    priority.text = "1.0"

    site = storage.get_site()

    if len(article.pages) > 1:
        languages = [language for language in site.allowed_languages if language in article.pages]
        add_language_variants(url, page_url, languages)


def add_language_variants(url, page_url, languages) -> None:
    for language in languages:
        ET.SubElement(
            url,
            "xhtml:link",
            rel="alternate",
            hreflang=language,
            href=page_url.to_language(language).url(),
        )



import os
import pathlib
from typing import Any

import jinja2

from brigid.core import logging
from brigid.jinja2_render.templates import get_jinjaglobals
from brigid.plugins.entities import FileInfo
from brigid.plugins.plugin import Plugin
from brigid.plugins.theme import jinjaglobals
from brigid.plugins.theme.settings import settings

logger = logging.get_module_logger()


class ThemePlugin(Plugin):
    def __init__(self, settings, **kwargs):
        super().__init__(**kwargs)
        self.settings = settings
        self._static_files_map = {}
        self._discover_static_files()

    def _discover_static_files(self) -> None:  # noqa: CCR001
        for path in [self.settings.static_redefined, self.settings.static_base]:
            if path is None:
                continue

            if not pathlib.Path(path).exists():
                logger.warning("static_path_not_found", path=path, plugin=self.slug)
                continue

            for root, _, files in os.walk(path):
                for file in files:
                    extension = file.split(".")[-1].lower()

                    if extension not in self.settings.static_extensions:
                        continue

                    media_type = self.settings.static_extensions[extension]

                    file_path = pathlib.Path(root) / file

                    url_path = str(pathlib.Path(file_path.relative_to(path)))

                    if file in self._static_files_map:
                        continue

                    self._static_files_map[file] = FileInfo(
                        sys_path=file_path, url_path=url_path, media_type=media_type
                    )

    def jinjaglobals(self) -> tuple[dict[str, Any], dict[str, Any]]:
        return get_jinjaglobals(jinjaglobals)

    def templates_loader(self) -> jinja2.BaseLoader:
        loaders = []

        if settings.templates_redefined:
            loaders.append(jinja2.FileSystemLoader(settings.templates_redefined, followlinks=True))

        loaders.append(jinja2.FileSystemLoader(settings.templates_base, followlinks=True))

        return jinja2.ChoiceLoader(loaders)

    def static_file_info(self, filename: str) -> FileInfo | None:
        return self._static_files_map.get(filename)

    def static_files(self) -> list[FileInfo]:
        return list(self._static_files_map.values())


plugin = ThemePlugin(slug="theme", settings=settings)



import os
import pathlib
from typing import Any

from markupsafe import Markup

from brigid.domain import request_context as d_request_context
from brigid.domain.urls import UrlsPlugin, UrlsRoot
from brigid.jinja2_render.utils import jinjafilter, jinjaglobal
from brigid.library import utils as l_utils
from brigid.library.entities import Page, PageSeriesInfo
from brigid.library.series import get_page_series_info
from brigid.markdown_render.markdown_render import render_page as markdown_render_page
from brigid.markdown_render.markdown_render import render_page_intro as markdown_render_page_intro
from brigid.markdown_render.markdown_render import render_text as markdown_render_text
from brigid.markdown_render.static_files import ImageInfo, files


@jinjafilter
def render_page(page: Page) -> str:
    return Markup(markdown_render_page(page=page).content)


@jinjafilter
def render_page_intro(page: Page) -> str:
    return Markup(markdown_render_page_intro(page=page).content)


@jinjafilter
def render_text(text: str) -> str:
    return Markup(markdown_render_text(text).content)


@jinjaglobal
def image_info(path: pathlib.Path) -> ImageInfo:
    return files.image_info(path)


@jinjaglobal
def root_url(language: str) -> UrlsRoot:
    return UrlsRoot(language=language)


@jinjaglobal
def plugin_url(plugin: str, language: str) -> UrlsPlugin:
    return UrlsPlugin(plugin, language=language)


@jinjaglobal
def test_marker(marker: str) -> str:

    if not os.environ.get("BRIGID_TESTS_RUNNING"):
        return ""

    return Markup(marker)


@jinjaglobal
def request_context_get(name: str) -> Any:
    return d_request_context.get(name)


@jinjaglobal
def page_series_info(page: Page) -> PageSeriesInfo:
    return get_page_series_info(page)


@jinjaglobal
def page_title(page: Page, short: bool = False) -> str:
    return l_utils.page_title(page=page, short=short)



from brigid.domain.urls import UrlsPost
from brigid.library.entities import Page
from brigid.library.storage import storage
from brigid.markdown_render.markdown_render import render_page, render_page_intro
from brigid.mcp.entities import Language, Post, PostInfo, PostMeta, RenderFormat, TagInfo


def create_post_meta(post: Page) -> PostMeta:
    article = storage.get_article(id=post.article_id)

    post_url = UrlsPost(language=post.language, slug=article.slug)

    return PostMeta(
        published_at=post.published_at,
        language=post.language,
        translated_into=set(article.pages.keys()),
        slug=article.slug,
        seo_description=post.seo_description,
        seo_image=post.seo_image,
        tags=create_tag_infos(post.language, {tag: 1 for tag in post.tags}),
        series=post.series,
        type=storage.get_article(id=post.article_id).type,
        http_url=post_url.url(),
    )


def create_post_info(post: Page, render_format: RenderFormat) -> PostInfo:
    meta = create_post_meta(post)

    match render_format:
        case RenderFormat.markdown:
            intro_body = post.intro
        case RenderFormat.html:
            render_context = render_page_intro(post)
            assert render_context.content
            intro_body = render_context.content
        case _:
            raise ValueError(f"Unsupported render format: {render_format}")

    return PostInfo(
        meta=meta, title=post.title, intro_format=render_format, intro_body=intro_body, has_more=post.has_more
    )


def create_post(post: Page, render_format: RenderFormat) -> Post:
    meta = create_post_meta(post)

    match render_format:
        case RenderFormat.markdown:
            body = post.body
        case RenderFormat.html:
            render_context = render_page(post)
            assert render_context.content
            body = render_context.content
        case _:
            raise ValueError(f"Unsupported render format: {render_format}")

    return Post(meta=meta, title=post.title, body_format=render_format, body=body)


def create_tag_infos(language: Language, tag_count: dict[str, int]) -> list[TagInfo]:
    site = storage.get_site()

    return [
        TagInfo(tag=tag, name=site.languages[language].tags_translations[tag], count=count)
        for tag, count in tag_count.items()
    ]



from unittest import mock

from brigid.domain.entities import Environment
from brigid.library.storage import storage


class TestSite:

    def test_url_choice(self) -> None:
        site = storage.get_site()

        # check default values
        assert str(site.prod_url) == "https://example.com/"
        assert str(site.local_url) == "http://0.0.0.0:8000/"

        del site.url

        with mock.patch("brigid.application.settings.settings.environment", Environment.prod):
            assert site.url == "https://example.com"

        del site.url

        with mock.patch("brigid.application.settings.settings.environment", Environment.local):
            assert site.url == "http://0.0.0.0:8000"



import fastapi
import pytest
from fastapi.testclient import TestClient

from brigid.domain import request_context

############################################################################
# ATTENTION: this tests do not cover some cases of request_context usage
#            because it setupped automatically in conftest.py for all tests.
#            In the future we can move this fixture from the global scope to the test level.
############################################################################


class TestFavicon:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        response = client.get("/favicon.ico")
        assert response.status_code == 200
        assert response.headers["content-type"] == "image/x-icon"


class TestSiteMap:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        response = client.get("/sitemap.xml")
        assert response.status_code == 200
        assert response.headers["content-type"] == "application/xml; charset=utf-8"


class TestPluginStatic:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        response = client.get("/static/plugins/theme/main.css")
        assert response.status_code == 200
        assert response.headers["content-type"] == "text/css; charset=utf-8"


class TestStaticFile:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        response = client.get("/static/posts/galery-types/images/image-1.jpg")
        assert response.status_code == 200
        assert response.headers["content-type"] == "image/jpeg"


class TestFeedAtom:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        for language in request_context.get("storage").get_site().allowed_languages:
            response = client.get(f"/{language}/feeds/atom")
            assert response.status_code == 200
            assert response.headers["content-type"] == "application/atom+xml; charset=utf-8"


class TestRobots:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        response = client.get("/robots.txt")
        assert response.status_code == 200
        assert response.headers["content-type"] == "text/plain; charset=utf-8"

    @pytest.mark.asyncio
    async def test_content(self, client: TestClient) -> None:
        expected_content = """
User-agent: *
Sitemap: http://0.0.0.0:8000/sitemap.xml
Disallow: /en/tags/
Disallow: /ru/tags/
        """

        response = client.get("/robots.txt")
        assert response.text == expected_content.strip()


class TestError:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        with pytest.raises(ZeroDivisionError):
            client.get("/test-error")

    @pytest.mark.asyncio
    async def test_exception_processing_works(self, app: fastapi.FastAPI) -> None:
        client = TestClient(app, raise_server_exceptions=False, follow_redirects=False)
        response = client.get("/test-error")
        assert response.status_code == 500


class TestRoot:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        response = client.get("/")
        assert response.status_code == 302
        assert response.headers["location"] == "http://0.0.0.0:8000/en"


class TestIndexRoot:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        for language in request_context.get("storage").get_site().allowed_languages:
            response = client.get(f"/{language}")
            assert response.status_code == 200
            assert response.headers["content-type"] == "text/html; charset=utf-8"


class TestIndexRootWithEmptyFilter:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        for language in request_context.get("storage").get_site().allowed_languages:
            response = client.get(f"/{language}/tags")
            assert response.status_code == 301
            assert response.headers["location"] == f"http://0.0.0.0:8000/{language}"


class TestIndexWithFilter:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        for language in request_context.get("storage").get_site().allowed_languages:
            response = client.get(f"/{language}/tags/example/-wide")
            assert response.status_code == 200
            assert response.headers["content-type"] == "text/html; charset=utf-8"


class TestPost:

    @pytest.mark.asyncio
    async def test_works(self, client: TestClient) -> None:
        for language in request_context.get("storage").get_site().allowed_languages:
            response = client.get(f"/{language}/posts/post-in-two-languages")
            assert response.status_code == 200
            assert response.headers["content-type"] == "text/html; charset=utf-8"



from typing import Any

from brigid.jinja2_render.templates import get_jinjaglobals
from brigid.plugins.i18n import jinjaglobals
from brigid.plugins.plugin import Plugin


class I18nPlugin(Plugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def jinjaglobals(self) -> tuple[dict[str, Any], dict[str, Any]]:
        return get_jinjaglobals(jinjaglobals)


plugin = I18nPlugin(slug="i18n")



import enum
from typing import Any

import pydantic

from brigid.core.entities import BaseEntity


class APIStatuses(str, enum.Enum):
    success = "success"
    error = "error"


class APIRequest(BaseEntity):
    model_config = pydantic.ConfigDict(title="Request Body")


class APISuccess(BaseEntity):
    status: APIStatuses = APIStatuses.success

    model_config = pydantic.ConfigDict(title="Response Body")


class APIError(BaseEntity):
    status: APIStatuses = APIStatuses.error
    code: str
    message: str = "Unknown error"
    data: dict[str, Any] | None = None

    model_config = pydantic.ConfigDict(title="API Error")



import enum
import re
import xml.etree.ElementTree as etree  # noqa: S405
from typing import Any

from markdown.inlinepatterns import LINK_RE as INTERNAL_LINK_RE
from markdown.inlinepatterns import LinkInlineProcessor

from brigid.domain.urls import UrlsFeedsAtom, UrlsPost, UrlsTags, normalize_url
from brigid.library.storage import storage
from brigid.markdown_render.context import render_context


class Option(enum.StrEnum):
    choose_nearest_language = "choose-nearest-language"
    language = "lang"
    language_title = "lang-title"


def extract_options(parts: list[str]) -> tuple[str, list[str], dict[Option, Any]]:  # noqa: CCR001
    options: dict[Option, Any] = {}

    tail = []

    for part in parts[1:]:
        if not part.startswith("@"):
            tail.append(part)
            continue

        if part[1:] == Option.choose_nearest_language:
            options[Option.choose_nearest_language] = True
            continue

        if part[1:].startswith(Option.language + "="):
            options[Option.language] = part[1 + len(Option.language) + 1 :].strip()
            continue

        if part[1:].startswith(Option.language_title + "="):
            options[Option.language_title] = part[1 + len(Option.language_title) + 1 :].strip()
            continue

        raise ValueError(f"Unknown option in local link: {part}")

    return parts[0], tail, options


class InternalLinkInlineProcessor(LinkInlineProcessor):
    RE_LINK = re.compile(r"\{\s*(.*?)\s*\}", re.DOTALL | re.UNICODE)

    # TODO: return errors instead of result
    def handleMatch(  # type: ignore  # noqa  # pylint: disable=all
        self, m: re.Match[str], data: str
    ) -> tuple[etree.Element | None, int | None, int | None]:
        from brigid.library.connectivity import connectivity

        result = super().handleMatch(m, data)

        if result[0] is None:
            return result  # type: ignore

        assert not isinstance(result[0], str)

        href = result[0].get("href")

        if href is None:
            return result  # type: ignore

        site = storage.get_site()

        context = render_context.get()

        parts = href.split(":")

        if len(parts) < 2:
            context.add_error(failed_text=data, message="Invalid local link format")
            return result  # type: ignore

        try:
            # TODO: move here all parsing?
            link_type, link_tail, options = extract_options(parts)
        except ValueError as e:
            context.add_error(failed_text=data, message=str(e))
            return result  # type: ignore

        new_href = None

        # TODO: support specifying language to easier give cross-language links

        if link_type == "post":
            slug = link_tail[0]

            if not storage.has_article(slug=slug):
                context = render_context.get()

                context.add_error(
                    failed_text=data,
                    message="Non-existing page slug specified for local link",
                )
                return result  # type: ignore

            article = storage.get_article(slug=slug)

            if options.get(Option.language) is not None:
                link_language = options[Option.language]

            elif options.get(Option.choose_nearest_language, False):
                link_language = article.first_language(
                    context.page.language, site.default_language, *site.allowed_languages
                )
            else:
                link_language = context.page.language

            if link_language is None:
                context.add_error(
                    failed_text=data,
                    message="Article does not have page in the current language",
                )
                return result  # type: ignore

            if options.get(Option.language_title) is not None:
                link_title_language = options[Option.language_title]
            else:
                link_title_language = link_language

            target_page_id = article.pages[link_language]
            storage.get_page(target_page_id)

            target_tile_page_id = article.pages[link_title_language]
            target_title_page = storage.get_page(target_tile_page_id)

            if link_language == context.page.language:
                # we track connection only for the same language
                # otherwise the "similar" section will be a mess
                connectivity.add_connection(
                    target_page_id=target_page_id,
                    reference_page_id=context.page.id,
                )

            new_href = UrlsPost(language=link_language, slug=slug).url()

            if not result[0].text:
                result[0].text = target_title_page.title

            if link_language != context.page.language or link_title_language != link_language:
                # add `[language]` to the text in link
                language_element = etree.Element("span")
                language_element.text = f" [{link_language}]"
                language_element.set("class", "brigid-link-language")
                result[0].insert(0, language_element)

        elif link_type == "tags":

            required = set()
            excluded = set()

            for tag in link_tail:
                normalized_tag = tag[1:] if tag[0] == "-" else tag

                if normalized_tag not in site.languages[context.page.language].tags_translations:
                    context.add_error(failed_text=data, message="One of tags does not exist")
                    return result  # type: ignore

                if tag[0] != "-":
                    required.add(normalized_tag)
                else:
                    excluded.add(normalized_tag)

            new_href = UrlsTags(
                page=1,
                required_tags=required,
                excluded_tags=excluded,
                language=context.page.language,
            ).url()

        elif link_type == "absolute":

            if link_tail[0][0] != "/":
                context.add_error(failed_text=data, message="Absolute link should start with /")
                return result  # type: ignore

            new_href = normalize_url(f"{site.url}/{link_tail[0]}")

        elif link_type == "feed":
            new_href = UrlsFeedsAtom(language=context.page.language).url()

        elif link_type == "static":
            post_url = UrlsPost(language=context.page.language, slug=context.article.slug)

            new_href = post_url.file_url(link_tail[0])

        else:
            context.add_error(failed_text=data, message="Unknown local link format")
            return result  # type: ignore

        # set full url for rendering in feeds and other places
        result[0].set("href", new_href)

        return result  # type: ignore

    def getLink(self, data: str, index: int) -> tuple[str, str | None, int, bool]:
        match = self.RE_LINK.match(data, index)

        if match is None:
            return "", None, -1, False

        href = match.group(1)

        return href, None, match.end(), True


__all__ = ["InternalLinkInlineProcessor", "INTERNAL_LINK_RE"]



import contextlib
import contextvars
import enum
from typing import Any, TypedDict

# TODO: these imports may cause circular dependencies, fix them
from brigid.domain.urls import UrlsBase
from brigid.library.storage import Storage


class Variable(enum.StrEnum):
    language = "language"
    url = "url"
    storage = "storage"


class RequestContextType(TypedDict):
    language: str
    url: UrlsBase
    storage: Storage


request_context: contextvars.ContextVar[RequestContextType] = contextvars.ContextVar("request_context")


@contextlib.contextmanager
def init():
    token = request_context.set({})

    try:
        yield
    finally:
        request_context.reset(token)


def set(name: str, value: Any):
    if name not in Variable:
        raise NotImplementedError(f"Variable {name} is not supported")

    context = request_context.get()

    if name in context:
        raise NotImplementedError(f"Variable {name} is already set in the request context")

    context[name] = value  # type: ignore


def get(name: str) -> Any:
    if name not in Variable:
        raise NotImplementedError(f"Variable {name} is not supported")

    context = request_context.get()

    if name not in context:
        raise NotImplementedError(f"Variable {name} is not set in the request context")

    return context[name]  # type: ignore


def current() -> RequestContextType:
    return request_context.get()



from typing import Any

from brigid.jinja2_render.utils import jinjafilter


@jinjafilter
def upper_first(text: str) -> str:
    return text[0].upper() + text[1:]


@jinjafilter
def to_str(value: Any) -> str:
    return str(value)



import datetime
import enum
import functools
import inspect
import logging
import uuid
from typing import Any, Callable, Iterable, Protocol, TypeVar

import pydantic_settings
import structlog
from sentry_sdk import capture_message
from structlog.contextvars import bound_contextvars


class Renderer(str, enum.Enum):
    console = "console"
    json = "json"


class Settings(pydantic_settings.BaseSettings):
    log_evel: str = "INFO"
    renderer: Renderer = Renderer.console

    model_config = pydantic_settings.SettingsConfigDict(
        env_nested_delimiter="__", env_file=".env", env_prefix="FFUN_", extra="allow"
    )

    @property
    def structlog_level(self) -> int:
        return getattr(logging, self.log_evel.upper())  # type: ignore


settings = Settings()


class LogProcessorType(Protocol):
    async def __call__(self, _: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
        pass


class Formatter:
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        raise NotImplementedError('You must implement "can_format" in child class')

    def format(self, value: Any) -> Any:
        raise NotImplementedError('You must implement "format" in child class')


class DateFormatter(Formatter):
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        return isinstance(value, (datetime.date, datetime.datetime))

    def format(self, value: Any) -> Any:
        return value.isoformat()


class UUIDFormatter(Formatter):
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        return isinstance(value, uuid.UUID)

    def format(self, value: Any) -> Any:
        return str(value)


class EnumFormatter(Formatter):
    __slots__ = ()

    def can_format(self, value: Any) -> bool:
        return isinstance(value, enum.Enum)

    def format(self, value: Any) -> Any:
        return value.name


class ProcessorFormatter:
    __slots__ = ("_formatters",)

    def __init__(self, formatters: Iterable[Formatter]) -> None:
        self._formatters = list(formatters)

    def __call__(self, _: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
        for key, value in event_dict.items():
            for formatter in self._formatters:
                if formatter.can_format(value):
                    event_dict[key] = formatter.format(value)
                    break

        return event_dict


def info_extracter(_: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
    replaced = False

    for key, value in list(event_dict.items()):
        if not hasattr(value, "log_info"):
            continue

        for k, v in value.log_info().items():
            event_dict[f"{key}_{k}"] = v

        del event_dict[key]

        replaced = True

    if not replaced:
        return event_dict

    return info_extracter(_, __, event_dict)


def create_formatter() -> ProcessorFormatter:
    formatters = [DateFormatter(), UUIDFormatter(), EnumFormatter()]
    return ProcessorFormatter(formatters)


def log_errors_to_sentry(_: Any, __: Any, event_dict: dict[str, Any]) -> dict[str, Any]:
    if event_dict.get("sentry_skip"):
        return event_dict

    if event_dict.get("level", "").upper() != "ERROR":
        return event_dict

    capture_message(event_dict["event"])

    return event_dict


def processors_list(use_sentry: bool) -> list[LogProcessorType]:
    sentry_processor = None

    if use_sentry:
        sentry_processor = log_errors_to_sentry

    processors_list = [
        structlog.contextvars.merge_contextvars,
        info_extracter,
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        sentry_processor,
        structlog.processors.TimeStamper(fmt="ISO", utc=True, key="timestamp"),
        create_formatter(),
        (structlog.dev.ConsoleRenderer() if settings.renderer == Renderer.console else None),
        (structlog.processors.JSONRenderer() if settings.renderer == Renderer.json else None),
    ]

    return [p for p in processors_list if p is not None]  # type: ignore


def initialize(use_sentry: bool) -> None:
    structlog.configure(
        processors=processors_list(use_sentry=use_sentry),  # type: ignore
        wrapper_class=structlog.make_filtering_bound_logger(settings.structlog_level),
        context_class=dict,
        logger_factory=structlog.PrintLoggerFactory(),
        cache_logger_on_first_use=True,
    )


def get_module_logger() -> structlog.stdlib.BoundLogger:
    caller_frame = inspect.currentframe().f_back  # type: ignore
    module = inspect.getmodule(caller_frame)
    return structlog.get_logger(module=module.__name__)  # type: ignore


FUNC = TypeVar("FUNC", bound=Callable[..., Any])


def bound_function(skip: Iterable[str] = ()) -> Callable[[FUNC], FUNC]:
    def wrapper(func: FUNC) -> FUNC:
        @functools.wraps(func)
        def wrapped(**kwargs: Any) -> Any:
            with bound_contextvars(**{k: v for k, v in kwargs.items() if k not in skip}):
                return func(**kwargs)

        @functools.wraps(func)
        async def async_wrapped(**kwargs: Any) -> Any:
            with bound_contextvars(**{k: v for k, v in kwargs.items() if k not in skip}):
                return await func(**kwargs)

        if inspect.iscoroutinefunction(func):
            return async_wrapped  # type: ignore

        return wrapped  # type: ignore

    return wrapper



from typing import Any

import pydantic
from pymdownx.blocks import BlocksExtension

from brigid.core.entities import BaseEntity
from brigid.markdown_render.processors.toml_block import TomlBlock


class CollectionModel(BaseEntity):
    id: str
    css_class: str | None = None

    model_config = pydantic.ConfigDict(frozen=False)


class CollectionBlock(TomlBlock):
    NAME = "brigid-collection"
    models = CollectionModel
    root_tag = "div"
    template: str = "theme/blocks/collection.html.j2"

    def root_css_classes(self, data: Any) -> list[str]:
        classes = ["brigid-collection", data.css_class]
        return [x for x in classes if x]


class CollectionBlockExtension(BlocksExtension):

    def extendMarkdownBlocks(self, md, block_mgr):
        block_mgr.register(CollectionBlock, self.getConfigs())



from pymdownx import snippets

from brigid.markdown_render.context import render_context


class BrigidSnippetPreprocessor(snippets.SnippetPreprocessor):

    @property
    def base_path(self):
        context = render_context.get()
        return [str(context.article.path.parent)]

    @base_path.setter
    def base_path(self, value):
        # do not allow to change base_path
        pass


class SnippetExtension(snippets.SnippetExtension):

    def extendMarkdown(self, md):
        self.md = md
        md.registerExtension(self)
        config = self.getConfigs()
        snippet = BrigidSnippetPreprocessor(config, md)
        md.preprocessors.register(snippet, "snippet", 32)
